package mon.battle

import scala.util.Random

import mon.stat._
import mon.stat.Statistic._
import mon.util.Fraction

class BattleEngine(val player: Party, val enemy: Party)(implicit val rng: Random) {

  def play() = {
    while (player.canFight && enemy.canFight) {
      playTurn()
    }
  }
  
  def playTurn() = {
    val playerMove = player.pollAction
    val playerTarget = player.pollTarget(playerMove, player.lead, enemy)
    val enemyMove = enemy.pollAction
    val enemyTarget = enemy.pollTarget(enemyMove, enemy.lead, player)
    val actions = Seq(Action(player.lead, playerMove, playerTarget), Action(enemy.lead, enemyMove, enemyTarget))
    val queue = rng.shuffle(actions).sorted // Shuffle to randomize in the event of a tie
    queue.foreach(useMove)
    val eotQueue = Seq(player.lead, enemy.lead).sortBy(_(Speed))
    eotQueue.foreach(mon => mon.status.map(_.onResidual.map(_(mon))))
    
    if (!player.lead.isAlive) {
      if (player.canFight) {
      val replace = player.pollReplacement(enemy) 
      player.swapIn(replace)
      } else {
        println(s"${player.trainer} has lost!")
      }
    }
    if (!enemy.lead.isAlive) {
      if (enemy.canFight) {
        val replace = enemy.pollReplacement(player)
        enemy.swapIn(replace)
      } else {
        println(s"${enemy.trainer} has lost!")
      }
    }
    
    println(s"${player.lead}(${player.lead.hp}/${player.lead(Hp)})")
    println(s"${enemy.lead}(${enemy.lead.hp}/${enemy.lead(Hp)})")
  }

  def useMove(action: Action) = {
    val user = action.user
    val move = action.move
    val target = action.target
    if (user.isAlive) {
      println(s"$user used $move.")
      if (attackRoll(user, move, target)) {
        if (move.pow > 0 || move.powCallback != null) {
          val dmg = damageRoll(user, move, target)
          val actualDmg = target.takeDamage(dmg)
          if (dmg == actualDmg) {
            println(s"$target takes $actualDmg damage!")
          } else {
            println(s"$target takes $actualDmg (+${dmg - actualDmg} overkill) damage!")
          }
          if (move.drain > 0.frac) {
            val healing = Math.max(move.drain * actualDmg, 1)
            user.recoverDamage(healing)
            println(s"$user recovered $healing damage.")
          } else if (move.drain < 0.frac) {
            val recoil = Math.max(-move.drain * actualDmg, 1)
            user.takeDamage(recoil)
            println(s"$user took $recoil damage from recoil!")
          }
          if (!user.isAlive) {
            println(s"$user fainted!")
          }
          if (!target.isAlive) {
            println(s"$target fainted!")
          }
        }
        applyBoosts(target, move.boosts)
        applyStatus(target, move.status)
        if (move.selfEffect != null) {
          applyEffect(user, move.selfEffect)
        }
        if (move.secondary != null && rng.chance(move.secondary.chance.%%)) {
          applyEffect(target, move.secondary)
        }

        // TODO : Support moves
        // TODO : Multiparty
      } else {
        println("Missed!")
      }
    }
  }

  def attackRoll(user: Monster, move: Move, target: Monster) = {
    if (move.accuracy == 0) {
      true
    } else {
      val acc = user.boosts.getOrElse(Accuracy, 0) - target.boosts.getOrElse(Evasion, 0)
      val mod = if (acc > 0) Fraction(3 + acc, 3) else Fraction(3, 3 - acc)
      val chance = move.accuracy.%% * mod
      rng.chance(chance)
    }
  }

  def damageRoll(user: Monster, move: Move, target: Monster) = {
    val atkStat = if (move.mvType == MoveType.Physical) PAtk else MAtk
    val defStat = if (move.mvType == MoveType.Physical) PDef else MDef
    // TODO : Fixed damage
    val pow = if (move.powCallback != null) move.powCallback(user, target) else move.pow
    val baseDmg = (2 * user.level / 5 + 2) * pow * user(atkStat) / (target(defStat) * 50) + 2
    val effectiveness = target.effectiveness(move.element)
    val multiplier = effectiveness * critMultiplier(user, move, target)
    if (effectiveness > 1.0) {
      println("It's super effective!")
    } else if (effectiveness < 1.0) {
      println("It's not very effective.")
    }
    val maxDmg = (baseDmg * multiplier).toInt
    val minDmg = (17 \ 20) * maxDmg
    rng.nextInt(minDmg, maxDmg)
  }
  
  def critMultiplier(user : Monster, move : Move, target: Monster) = {
    // Percentage chance is different from Pokemon
    val stage = move.crit
    val chance = (1 << stage) \ 16 // Doubles per stage
    if (rng.chance(chance)) {
      println("A critical hit!")
      1.5
    } else {
      1.0
    }
  }

  def applyEffect(target: Monster, effect: Secondary) {
    applyBoosts(target, effect.boosts)
    applyStatus(target, effect.status)
  }

  def applyStatus(target: Monster, status: Status) {
    if (target.isAlive) {
      if (status != null && target.status == None) {
        // apply status
        target.status = Some(status)
        status.onStart.map(_(target))
      }
    }
  }

  def applyBoosts(target: Monster, boosts: Map[Stat, Int]) {
    if (target.isAlive) {
      boosts.foreach {
        case (s, b) => {
          target.applyBoost(s, b)
          if (b > 0) {
            println(s"$target's $s rose!")
          } else if (b < 0) {
            println(s"$target's $s fell!")
          }
        }
      }
    }
  }
}